Define various atomic types: variables, typing of variables, functions, key
inverses, timestamps.

>module Atoms (
>  -- variables
>  VarName, AgentId,
>  -- types of variables
>  TypeName, VarTypeList, ActVarTypeList, 
>  findtype1, findtype1T, allOfType, allTypes,
>  Status(Normal,Foreground,UnknownBackground,KnownBackground,Dummy,Error_Status),
>  -- functions
>  FnType(Symb, Defed, HashFunction), 
>  FnTypeList, isFn, isHashFunction, findFnType, findFnType1, 
>  findFnDomType, findFnRanType, 
>  -- types
>  isAType,
>  -- inverses
>  InverseKeyList, inverse1, -- inverse2, 
>  -- timestamps
>  TimeStampInfo(NotUsed, MRTUsed, TimeStampsUsed), 
>  DataTypeDef, DataTypeDefs, dtAtoms, dtConss, notUsed,
>  tryin)

>where

>import Useful
>import Maybe1

=========================================================================

Variable names, agent identities

>type VarName = String
>type AgentId = String

=========================================================================
Typing of variables

>type TypeName = String

>type VarTypeList = [(VarName, TypeName)]

>data Status =	Foreground | 
>		KnownBackground | UnknownBackground |
>		Normal | Dummy | Error_Status
>		deriving(Eq, Show)

>type ActVarTypeList = [(VarName, TypeName, Status)]

>findtype :: VarTypeList -> VarName -> Maybe_ TypeName
>findtype tl v 
>  | matches == []       = Error ("Undeclared variable: " ++ v ++ "\n")
>  | length matches > 1  = Error ("Multiply declared variable: " ++ v ++ "\n")
>  | length matches == 1 = Yes (hd matches)
>      where matches = [t | (v',t) <- tl {- ++ encrypttl -} , v==v']

>findtype1 :: VarTypeList -> VarName -> TypeName
>findtype1 tl v = 
>  hd ([t | (v',t) <- tl++nowtl, v==v' || v==v'++"_X_"] ++ ["_Internal"])

The comparison with v'++"_X_" is a hack used for generating new type names
where the same field appears in both senders-extras and receivers-extras.  It
is used in Channels.lhs and RenamedSystem.lhs

>nowtl = [("now", "TimeStamp")]

>allOfType tl t = sortremdups [v | (v,t') <- tl {- ++ encrypttl -}, t == t']

>allTypes tl v = [t | (v',t) <- tl, v==v']

Return type, but substitute TS for TimeStamp

>findtype1T tl v = 
>  let t1 = findtype1 tl v 
>  in if t1=="TimeStamp" then "TS" else t1

>-- encrypttl = [] -- ("Encrypt", "ENCRYPT"), ("EndEncrypt", "ENDENCRYPT")]

=========================================================================
Functions

>data FnType = Symb [TypeName] TypeName |    -- symbolic functions
>              Defed [TypeName] TypeName |   -- functions with explicit def
>              HashFunction String              -- public hash functions
>              deriving (Eq, Show)
>type FnTypeList = [(VarName, FnType)]

>findFnType :: FnTypeList -> VarName -> Maybe_ FnType
>findFnType _ "addr" = Yes(Defed ["Agent"] "Message")
>findFnType fnts f 
>  | matches == []       = Error ("Undeclared function: " ++ f ++ (show fnts)++"\n")
>  | length matches > 1  = Error ("Multiply declared function: " ++ f ++ "\n")
>  | length matches == 1 = Yes (hd matches)
>      where matches = [t | (g,t) <- fnts, f == g]

>isHashFunction :: FnTypeList -> VarName -> Bool
>isHashFunction fnts f = tryin (findFnType1 fnts f) 

>tryin :: FnType -> Bool
>tryin (HashFunction x) = True
>tryin _ = False

>findFnType1 :: FnTypeList -> VarName -> FnType
>findFnType1 fnts f = let Yes t = findFnType fnts f in t

>isFn :: FnTypeList -> VarName -> Bool
>isFn fnts f = 
>  case findFnType fnts f of
>    Error _ -> False
>    Yes _ -> True

>findFnDomType :: FnTypeList -> VarName -> [TypeName]
>findFnDomType fnts f = 
>  let Yes ft = findFnType fnts f
>  in 
>  case ft of
>    Symb ats rt -> ats
>    Defed ats rt -> ats

>findFnRanType :: FnTypeList -> VarName -> TypeName
>findFnRanType fnts f = 
>  let Yes ft = findFnType fnts f
>  in 
>  case ft of
>    Symb ats rt -> rt
>    Defed ats rt -> rt


>isAType :: VarTypeList -> FnTypeList -> TypeName -> Bool
>isAType fvts fnts tn = 
>  let allTypes = map snd fvts++
>                 [t | (_,Symb _ t) <- fnts]++
>                 [t | (_,Defed _ t) <- fnts]
>  in member tn allTypes


=========================================================================

Inverses of keys

>type InverseKeyList = [(VarName, VarName)]

>inverse :: InverseKeyList -> VarName -> Maybe_ VarName

>inverse fiks v 
>  | length matches == 1   = Yes (hd matches)
>  | matches == []         = Error ("Inverse not found: " ++ v ++ "\n")
>  | length matches > 1    = Error ("Multiply defined inverse: " ++ v ++ "\n")
>  where matches = [v' | (v'',v') <- fiks, v'' == v] 
>                  ++ [v' | (v',v'') <- fiks, v'' == v, v' /= v'']

>inverse1 :: InverseKeyList -> VarName -> VarName
>inverse1 fiks v = body (inverse fiks v)

>{-
>inverse2 :: InverseKeyList -> Msg -> Msg
>inverse2 fiks (Atom k) = Atom (inverse1 fiks k)
>inverse2 fiks (Apply f as) = Apply (inverse1 fiks f) as
>inverse2 fiks (Forwd v m) = inverse2 fiks m
>inverse2 fiks (Undec m v) = inverse2 fiks m
>-}


=========================================================================
Time stamp information

>data TimeStampInfo = NotUsed | MRTUsed Int | TimeStampsUsed Int Int Int
>                     deriving (Eq,Show)

>notUsed=NotUsed

* NotUsed represents that time is not used at all
* MRTUsed rt represents that each run will take at most rt time units; 
should occur only when there are timed authentications or timed key compromise 
* TimeStampsUsed t0 t1 rt represents that timestamps will take the range 
t0..t1, and each run will take at most rt time units; should occur only when
timestamps are used 

=========================================================================

Datatype definitions

>type DataTypeDef = (TypeName, [Pattern], Int)
>type Pattern = (VarName, [VarName])
>type DataTypeDefs = [DataTypeDef]

Atomic values of datatype definition

>dtAtoms :: DataTypeDef -> [VarName]
>dtAtoms (tn, pats, n) = [cons | (cons,[]) <- pats]

Constructors of datatype definition

>dtConss :: DataTypeDef -> [VarName]
>dtConss (tn, pats, n) = [cons | (cons,_:_) <- pats]
