Perform various consistency checks for scripts using data independence 
techniques.

>module ConsistencyDI (
>  warningDI_authStale, checkFreshVarFlow
>  ) 

>where

>import Useful
>import Pprint
>import Atoms
>import Types
>import Accumulation

========================================================================
Consistency check for DI scripts:

Conditions Required:
	   
1. The environment must not introduce any DI variables in some message M
   which are fresh to the Receiver of M.

2. If the receiver of M is the environment, then the sender of M must not
introduce any fresh variables of some DI type in M.


>checkFreshVarFlow fvts dit protdesc =
>  let	 accum_init = accumulation_init fvts protdesc dit
>        isDi t = member t dit
>        accum_init' = [(nr, (s, filter isDi sl), (r, filter isDi rl)) |
>                         (nr, (s,sl), (r,rl)) <- accum_init]
>	 checkListS = [(nr,remdups(sl++rl)) | 
>			  (nr,(s,sl),(_,rl)) <- accum_init', 
>                         s==Environment, sl /= []]
>	 checkListR = [(nr,s,sl) | (nr,(Agent s,sl),(r,_)) <- accum_init,
>				  r==Environment, sl/=[]]
>  in
>	 concat
>	  ["Error: Message " ++ nr ++
>	   " should not introduce the fresh variables " ++ 
>	   commaConcat ls ++ "\n" ++ 
>	   "by the Environment (as Sender).\n\n" |
>		       (nr,ls) <- checkListS] ++
>
>	 concat
>	  ["Error: Message " ++ nr ++
>	   " should not introduce the fresh variables " ++ 
>	   commaConcat sl ++ "\n" ++ "by " ++ s ++
>	   "and send them to the Environment.\n\n" |
>		       (nr,s,sl) <- checkListR]


===================================================================
Consistency warning concerning the use of the Aliveness specification
combined with StaleKnowledge = True.  System using the DI techniques
with StaleKnowledge=True will, by definition of stale knowledge,
automatically refine the Aliveness Spec.

>warningDI_authStale stale auths = 
>  let
>	specs = [(r,c) | (r,c,Aliveness) <- auths] /= []
>  in
>	if (stale && specs) then
>	  "\n" ++ "WARNING:\n" ++
>	  "All Aliveness specifications will hold true for this script,\n"
>	  ++
>	  "using data independence, due to the use of stale knowledge.\n"
>	  ++
>	  "The stale knowledge implies that all combinations of agents\n"
>	  ++
>	  "have previously run the protocol.  Using the Aliveness spec.\n"
>	  ++
>	  "in this context does not really provide any useful analysis.\n\n"
>	else ""

