>module Messages (
>  -- messages
>  Msg(Atom, Encrypt, Sq, Xor, Undec, Forwd, Apply), funname,
>  findtype2, atomsSend, atomsRec, atoms_all, atomsSend_accum,
>  atomsRec_accum, atoms, senderAtomFns, receiverAtomFns, fnApps,
>  receiverfnApps, VarField(Simple, Subst), fnPerCents,
>  receiverPerCents, noPerCents, receiverFields, receiverFields_Di,
>  senderFields, receiver2Fields, varFields, varFields2, senderAtoms,
>  senderTimeStamps, keys, anyOfTypes, receiverKeys, receiverMsg,
>  receiverMsg_Di, senderMsg, varMsg, patternMatchable, showMsg,
>  showSenderMsg, showReceiverMsg, showReceiverChanMsg, showFn,
>  showPatMsg, components1, typeSet, makegen, makegens, makeVarGens,
>  TaggedVarName(Percent, Free)
>  ) 

>where

>import Useful
>import Pprint
>import Atoms

=========================================================================
Protocol messages

>type Key = VarName

>data Msg = Atom VarName | Encrypt Msg [Msg] | Sq [Msg] | Xor Msg Msg |
>           Undec Msg VarName | Forwd VarName Msg | Apply VarName Msg
>           deriving (Eq, Ord, Show)

Undecryptable k ms enc represents a message {ms}{k} that the receiver cannot
decrypt, but which he stores in the variable enc.  Forward enc k ms represents
a message stored in enc which is to be forwarded and interpreted as being of
the form {ms}{k}.

function to produce the datatype constructor representing a symbolic function

>funname f = f++"__"

=====================================================================

Find type of a message, which should repesent a variable

>findtype2 :: VarTypeList -> FnTypeList -> Msg -> VarName
>findtype2 fvts fnts (Atom a) = findtype1 fvts a
>findtype2 fvts fnts (Apply f as) = findFnRanType fnts f
>findtype2 fvts fnts (Forwd v m) = findtype2 fvts fnts m
>findtype2 fvts fnts (Undec m v) = findtype2 fvts fnts m

=========================================================================

return atoms of message, sorted, with no duplicates

>atoms :: Msg -> [VarName]
>atoms (Atom v) = [v]
>atoms (Encrypt k ms) = sortremdupsconcat (atoms k : map atoms ms)
>atoms (Sq ms) = sortremdupsconcat (map atoms ms)
>atoms (Xor m m') = remdupsmerge (atoms m) (atoms m')
>atoms (Undec m v) = atoms m
>atoms (Forwd v m) = atoms m
>atoms (Apply f m) = atoms m


return atoms of message including m and v in the cases of m%v and v%m,
sorted, with no duplicates. (Used in TypeCheckDI.lhs only)

>atoms_all :: Msg -> [VarName]
>atoms_all (Atom v) = [v]
>atoms_all (Encrypt k ms) = sortremdupsconcat (atoms_all k : map atoms_all ms)
>atoms_all (Sq ms) = sortremdupsconcat (map atoms_all ms)
>atoms_all (Xor m m') = remdupsmerge (atoms_all m) (atoms_all m')
>atoms_all (Undec m v) = v : (atoms_all m)
>atoms_all (Forwd v m) = v : (atoms_all m)
>atoms_all (Apply f m) = atoms_all m

return atoms of message which are visible to the sender sorted, with
no duplicates.

>atomsSend :: Msg -> [VarName]
>atomsSend (Atom v) = [v]
>atomsSend (Encrypt k ms) = sortremdupsconcat (atomsSend k : map atomsSend ms)
>atomsSend (Sq ms) = sortremdupsconcat (map atomsSend ms)
>atomsSend (Xor m m') = remdupsmerge (atomsSend m) (atomsSend m')
>atomsSend (Undec m v) = atomsSend m
>atomsSend (Forwd v m) = [v]
>atomsSend (Apply f m) = atomsSend m

return atoms of message which are visible to the receiver, sorted,
with no duplicates.

>atomsRec :: Msg -> [VarName]
>atomsRec (Atom v) = [v]
>atomsRec (Encrypt k ms) = sortremdupsconcat (atomsRec k : map atomsRec ms)
>atomsRec (Sq ms) = sortremdupsconcat (map atomsRec ms)
>atomsRec (Xor m m') = remdupsmerge (atomsRec m) (atomsRec m')
>atomsRec (Undec m v) = [v]
>atomsRec (Forwd v m) = atomsRec m
>atomsRec (Apply f m) = atomsRec m

The function "atomsRec_acc is used for the DI work only.  Needed to
distinguish between fresh global variables and those local to a
%-expression.  

For example:
		Msg 1. A -> B : x
		Msg 2. B -> A : x % v

In this case, we need to store x as fresh in msg 1 AND x as fresh
in msg 2.  Without tagging these %-expressions, the x in msg 2 would
have been disgarded in the filtering process of the "accumulation"
function.

>data TaggedVarName = Percent VarName | Free VarName
>			deriving (Eq, Ord, Show)

>atomsSend_accum :: Msg -> [TaggedVarName]
>atomsSend_accum (Atom v) = [Free v]
>atomsSend_accum (Encrypt k ms) = sortremdupsconcat 
>				(atomsSend_accum k : map atomsSend_accum ms)
>atomsSend_accum (Sq ms) = sortremdupsconcat (map atomsSend_accum ms)
>atomsSend_accum (Xor m m') = remdupsmerge (atomsSend_accum m) 
>					(atomsSend_accum m')
>atomsSend_accum (Undec m v) = atomsSend_accum m
>atomsSend_accum (Forwd v m) = [Free v]
>atomsSend_accum (Apply f m) = atomsSend_accum m


>atomsRec_accum :: Msg -> [TaggedVarName]
>atomsRec_accum (Atom v) = [Free v]
>atomsRec_accum (Encrypt k ms) = sortremdupsconcat (atomsRec_accum k : 
>						   map atomsRec_accum ms)
>atomsRec_accum (Sq ms) = sortremdupsconcat (map atomsRec_accum ms)
>atomsRec_accum (Xor m m') = remdupsmerge (atomsRec_accum m)
>					 (atomsRec_accum m')
>atomsRec_accum (Undec m v) = [Percent v | v <- atoms m]
>atomsRec_accum (Forwd v m) = atomsRec_accum m
>atomsRec_accum (Apply f m) = atomsRec_accum m


return set of function applications sorted, with no duplicates

>fnApps :: Msg -> [(VarName, Msg)]
>fnApps (Atom v) = []
>fnApps (Encrypt k ms) = sortremdupsconcat (fnApps k : map fnApps ms)
>fnApps (Sq ms) = sortremdupsconcat (map fnApps ms)
>fnApps (Xor m m') = remdupsmerge (fnApps m) (fnApps m')
>fnApps (Undec m v) = fnApps m
>fnApps (Forwd v m) = fnApps m
>fnApps (Apply f m) = [(f, m)] 

Function applications as seen by receiver, excluding keys

>receiverfnApps :: Msg -> [(VarName, Msg)]
>receiverfnApps (Atom v) = []
>receiverfnApps (Encrypt k ms) = sortremdupsconcat (map receiverfnApps ms)
>receiverfnApps (Sq ms) = sortremdupsconcat (map receiverfnApps ms)
>receiverfnApps (Xor m m') = remdupsmerge (receiverfnApps m) (receiverfnApps m')
>receiverfnApps (Undec m v) = [] -- receiverfnApps m
>receiverfnApps (Forwd v m) = receiverfnApps m
>receiverfnApps (Apply f m) = [(f, m)] 

>data VarField = Simple VarName | Subst VarName Msg
>                deriving (Eq, Ord, Show)

Return function-varaible pairs used in %-terms

>fnPerCents :: Msg -> [(VarName, VarName)]
>fnPerCents (Atom _) = []
>fnPerCents (Encrypt k ms) = fnPerCents k ++ flatmap fnPerCents ms
>fnPerCents (Sq ms) = flatmap fnPerCents ms
>fnPerCents (Xor m m') = fnPerCents m ++ fnPerCents m'
>fnPerCents (Undec m v) = 
>  case m of
>    Apply f _ -> [(f,v)]
>    _ -> fnPerCents m
>fnPerCents (Forwd v m) = 
>  case m of
>    Apply f _ -> [(f,v)]
>    _ -> fnPerCents m
>fnPerCents (Apply _ _) = []

List of variables appearing on RHSs of % constructs

>receiverPerCents :: Msg -> [String]
>receiverPerCents (Atom _) = []
>receiverPerCents (Encrypt _ ms) = flatmap receiverPerCents ms
>receiverPerCents (Sq ms) = flatmap receiverPerCents ms
>receiverPerCents (Xor m m') = receiverPerCents m ++ receiverPerCents m'
>receiverPerCents (Undec m v) = [v]
>receiverPerCents (Forwd v m) = receiverPerCents m
>receiverPerCents (Apply _ m) = receiverPerCents m

Is a message free from %-s

>noPerCents :: Msg -> Bool
>noPerCents (Atom _) = True
>noPerCents (Encrypt k ms) = all noPerCents ms
>noPerCents (Sq ms) = all noPerCents ms
>noPerCents (Xor m m') = noPerCents m && noPerCents m'
>noPerCents (Undec _ _) = False
>noPerCents (Forwd _ _) = False
>noPerCents (Apply _ m) = noPerCents m


fnPerCents m

Return set of fields, as seen be receiver

>receiverFields :: Msg -> [VarField]
>receiverFields (Atom v) = [Simple v]
>receiverFields (Encrypt k ms) = sortremdupsconcat (map receiverFields ms)
>receiverFields (Sq ms) = sortremdupsconcat (map receiverFields ms)
>receiverFields (Xor m m') = 
>  remdupsmerge (receiverFields m) (receiverFields m')
>receiverFields (Undec m v) = [Subst v m]
>receiverFields (Forwd v m) = receiverFields m
>receiverFields (Apply _ _) = []

>receiverFields_Di :: Msg -> [VarField]
>receiverFields_Di (Atom v) = [Simple v]
>receiverFields_Di (Encrypt k ms) = 
>		sortremdupsconcat (map receiverFields_Di ms)
>receiverFields_Di (Sq ms) = sortremdupsconcat (map receiverFields_Di ms)
>receiverFields_Di (Xor m m') = 
>  remdupsmerge (receiverFields_Di m) (receiverFields_Di m')
>receiverFields_Di (Undec m v) = receiverFields_Di m
>receiverFields_Di (Forwd v m) = receiverFields_Di m
>receiverFields_Di (Apply _ _) = []

>receiver2Fields :: Msg -> [VarField]
>receiver2Fields (Atom v) = [Simple v]
>receiver2Fields (Encrypt k ms) = 
>  sortremdupsconcat (receiver2Fields k:map receiver2Fields ms)
>receiver2Fields (Sq ms) = sortremdupsconcat (map receiver2Fields ms)
>receiver2Fields (Xor m m') = 
>  remdupsmerge (receiver2Fields m) (receiver2Fields m')
>receiver2Fields (Undec m v) = [Subst v m]
>receiver2Fields (Forwd v m) = receiver2Fields m
>receiver2Fields (Apply f m) = receiver2Fields m

Return set of fields, including keys

>varFields :: Msg -> [VarField]
>varFields (Atom v) = [Simple v]
>varFields (Encrypt k ms) = sortremdupsconcat (varFields k : map varFields ms)
>varFields (Sq ms) = sortremdupsconcat (map varFields ms)
>varFields (Xor m m') = remdupsmerge (varFields m) (varFields m')
>varFields (Undec m v) = [Subst v m]
>varFields (Forwd v m) = [Subst v m]
>varFields (Apply f m) = varFields m 

Returns set of fields, as seen by Sender

>senderFields :: Msg -> [VarField]
>senderFields (Atom v) = [Simple v]
>senderFields (Encrypt k ms) = sortremdupsconcat 
>			         (senderFields k : map senderFields ms)
>senderFields (Sq ms) = sortremdupsconcat (map senderFields ms)
>senderFields (Xor m m') = remdupsmerge (senderFields m) (senderFields m')
>senderFields (Undec m v) = senderFields m
>senderFields (Forwd v m) = [Subst v m]
>senderFields (Apply f m) = senderFields m 


Same as varFields above except on a list of 'extra' variables

>varFields2 :: [String] -> [VarField]
>varFields2 [] = []
>varFields2 (x:xs) = Simple x:varFields2 xs

Does agent's identity id appear in message in pattern-matchable way?

>patternMatchable :: AgentId -> Msg -> Bool
>patternMatchable id (Atom v) = id == v
>patternMatchable id (Encrypt k ms) = any (patternMatchable id) ms
>patternMatchable id (Sq ms)  = any (patternMatchable id) ms
>patternMatchable id (Xor m m') = 
>  patternMatchable id m || patternMatchable id m'
>patternMatchable id (Undec m v) = False -- patternMatchable id m
>patternMatchable id (Forwd v m) = False -- patternMatchable id m
>patternMatchable id (Apply _ _) = False

Atoms, from sender's point of view

>senderAtoms :: Msg -> [VarName]
>senderAtoms (Atom v) = [v]
>senderAtoms (Encrypt k ms) = 
>  sortremdupsconcat (senderAtoms k : map senderAtoms ms)
>senderAtoms (Sq ms) = sortremdupsconcat (map senderAtoms ms)
>senderAtoms (Xor m m') = remdupsmerge (senderAtoms m) (senderAtoms m')
>senderAtoms (Undec m v) = senderAtoms m
>senderAtoms (Forwd v m) = [v]
>senderAtoms (Apply f m) = f : senderAtoms m

>senderAtomFns :: Msg -> [Msg]
>senderAtomFns (Atom v) = [Atom v]
>senderAtomFns (Encrypt k ms) = 
>  sortremdupsconcat (senderAtomFns k : map senderAtomFns ms)
>senderAtomFns (Sq ms) = sortremdupsconcat (map senderAtomFns ms)
>senderAtomFns (Xor m m') = remdupsmerge (senderAtomFns m) (senderAtomFns m')
>senderAtomFns (Undec m v) = senderAtomFns m
>senderAtomFns (Forwd v m) = [Atom v]
>senderAtomFns (Apply f as) = [Apply f as]

Now from receiver's point of view

>receiverAtomFns :: Msg -> [Msg]
>receiverAtomFns (Atom v) = [Atom v]
>receiverAtomFns (Encrypt k ms) = 
>  sortremdupsconcat (receiverAtomFns k : map receiverAtomFns ms)
>receiverAtomFns (Sq ms) = sortremdupsconcat (map receiverAtomFns ms)
>receiverAtomFns (Xor m m') = 
>  remdupsmerge (receiverAtomFns m) (receiverAtomFns m')
>receiverAtomFns (Undec m v) = [Atom v]
>receiverAtomFns (Forwd v m) = receiverAtomFns m
>receiverAtomFns (Apply f as) = [Apply f as]

List of timestamps appearing in message, from sender's point of view

>senderTimeStamps :: VarTypeList -> FnTypeList -> Msg -> [VarName]
>senderTimeStamps fvts fnts (Atom v) = 
>  if findtype1 fvts v == "TimeStamp" then [v] else []
>senderTimeStamps fvts fnts (Encrypt k ms) = 
>  sortremdupsconcat (map (senderTimeStamps fvts fnts) ms)
>senderTimeStamps fvts fnts (Sq ms) = 
>  sortremdupsconcat (map (senderTimeStamps fvts fnts) ms)
>senderTimeStamps fvts fnts (Xor m m') =
>  remdupsmerge (senderTimeStamps fvts fnts m) (senderTimeStamps fvts fnts m') 
>senderTimeStamps fvts fnts (Undec m v) = senderTimeStamps fvts fnts m
>senderTimeStamps fvts fnts (Forwd v m) = []
>senderTimeStamps fvts fnts (Apply f m) = 
>  if isHashFunction fnts f then senderTimeStamps fvts fnts m else [] 

return keys used for encrypting parts of message

>keys (Atom _) = []
>keys (Encrypt k ms) = sortremdupsconcat ([k] : map keys ms)
>keys (Sq ms) = sortremdupsconcat (map keys ms)
>keys (Xor m m') = remdupsmerge (keys m) (keys m')
>keys (Undec m v) = keys m
>keys (Forwd v m) = keys m
>keys (Apply _ _) = []

Test whether message uses any field of a type from ts

>anyOfTypes :: VarTypeList -> FnTypeList -> [TypeName] -> Msg -> Bool
>anyOfTypes fvts fnts ts (Atom a) = member (findtype1 fvts a) ts
>anyOfTypes fvts fnts ts (Encrypt k ms) =
>  anyOfTypes fvts fnts ts k || any (anyOfTypes fvts fnts ts) ms
>anyOfTypes fvts fnts ts (Sq ms) = any (anyOfTypes fvts fnts ts) ms
>anyOfTypes fvts fnts ts (Xor m m') = 
>  anyOfTypes fvts fnts ts m || anyOfTypes fvts fnts ts m'
>anyOfTypes fvts fnts ts (Undec m v) = anyOfTypes fvts fnts ts m
>anyOfTypes fvts fnts ts (Forwd v m) = anyOfTypes fvts fnts ts m
>anyOfTypes fvts fnts ts (Apply f m) = 
>  not (isHashFunction fnts f) && member (findFnRanType fnts f) ts 
>  || anyOfTypes fvts fnts ts m


>receiverKeys (Atom _) = []
>receiverKeys (Encrypt k ms) = sortremdupsconcat ([k] : map receiverKeys ms)
>receiverKeys (Sq ms) = sortremdupsconcat (map receiverKeys ms)
>receiverKeys (Xor m m') = remdupsmerge (receiverKeys m) (receiverKeys m')
>receiverKeys (Undec m v) = []
>receiverKeys (Forwd v m) = receiverKeys m
>receiverKeys (Apply _ _) = []

Message as experienced by sender

>senderMsg :: Msg -> Msg
>senderMsg (Atom v) = Atom v
>senderMsg (Encrypt k ms) = Encrypt (senderMsg k) (map senderMsg ms)
>senderMsg (Sq ms) = Sq (map senderMsg ms)
>senderMsg (Xor m m') = Xor (senderMsg m) (senderMsg m')
>senderMsg (Undec m v) = senderMsg m
>senderMsg (Forwd v m) = Atom v
>senderMsg (Apply f as) = Apply f as

Message as experienced by receiver

>receiverMsg :: Msg -> Msg
>receiverMsg (Atom v) = Atom v
>receiverMsg (Encrypt k ms) = Encrypt (receiverMsg k) (map receiverMsg ms)
>receiverMsg (Sq ms) = Sq (map receiverMsg ms)
>receiverMsg (Xor m m') = Xor (receiverMsg m) (receiverMsg m')
>receiverMsg (Undec m v) = Atom v
>receiverMsg (Forwd v m) = receiverMsg m
>receiverMsg (Apply f as) = Apply f as

Used solely for the purpose of constructing the "get" message sets in
"defMsgSets" (in DI work)

>receiverMsg_Di :: Msg -> Msg
>receiverMsg_Di (Atom v) = Atom v
>receiverMsg_Di (Encrypt k ms) = Encrypt (receiverMsg_Di k) 
>					 (map receiverMsg_Di ms)
>receiverMsg_Di (Sq ms) = Sq (map receiverMsg_Di ms)
>receiverMsg_Di (Xor m m') = Xor (receiverMsg_Di m) (receiverMsg_Di m')
>receiverMsg_Di (Undec m v) = receiverMsg_Di m
>receiverMsg_Di (Forwd v m) = receiverMsg_Di m
>receiverMsg_Di (Apply f as) = Apply f as

Message with variables in % cases

>varMsg :: Msg -> Msg
>varMsg (Atom v) = Atom v
>varMsg (Encrypt k ms) = Encrypt (varMsg k) (map varMsg ms)
>varMsg (Sq ms) = Sq (map varMsg ms)
>varMsg (Xor m m') = Xor (varMsg m) (varMsg m')
>varMsg (Undec m v) = Atom v
>varMsg (Forwd v m) = Atom v
>varMsg (Apply f m) = Apply f (varMsg m)

============================= Printing functions

Print function application or hash function; argument showArgsFn is the
function to be used to print each argument.

>showFn fnts showArgsFn (Apply f m) = 
>  if isHashFunction fnts f 
>  then "Hash.("++f++", <"++showArgs showArgsFn m++">)"
>  else f++"("++showArgs showArgsFn m++")"

>showArgs showArgsFn (Sq ms) = commaConcat (map showArgsFn ms)
>showArgs showArgsFn m = showArgsFn m

Return string representing message.  Old version does not add tags to
timestamps.

>{-
>showMsg :: FnTypeList -> Msg -> String
>showMsg fnts (Atom v) = v
>showMsg fnts (Encrypt k ms) = 
>  "Encrypt.(" ++ showMsg fnts k ++ ", <" ++ 
>  commaConcat (map (showMsg fnts) ms) ++ ">)"
>showMsg fnts (Sq ms) = "Sq.<" ++ commaConcat (map (showMsg fnts) ms) ++ ">"
>showMsg fnts (Xor m m') =  
>  "Xor.(" ++ showMsg fnts m ++ ", " ++ showMsg fnts m' ++ ")"
>showMsg fnts (Undec m v) = showMsg fnts m
>showMsg fnts (Forwd v m) = showMsg fnts m
>showMsg fnts (Apply f m) = showFn fnts (showMsg fnts) (Apply f m)
>-}

Return string representing message, tagging timestamps as such

>showMsg :: VarTypeList -> FnTypeList -> Msg -> String
>showMsg fvts fnts (Atom v) = 
>  if findtype1 fvts v == "TimeStamp" then "Timestamp."++v else v
>showMsg fvts fnts (Encrypt k ms) = 
>  "Encrypt.(" ++ showMsg fvts fnts k ++ ", <" ++ 
>  commaConcat (map (showMsg fvts fnts) ms) ++ ">)"
>showMsg fvts fnts (Sq ms) = 
>  "Sq.<" ++ commaConcat (map (showMsg fvts fnts) ms) ++ ">"
>showMsg fvts fnts (Xor m m') =  
>  "Xor.(" ++ showMsg fvts fnts m ++ ", " ++ showMsg fvts fnts m' ++ ")"
>showMsg fvts fnts (Undec m v) = showMsg fvts fnts m
>showMsg fvts fnts (Forwd v m) = showMsg fvts fnts m
>showMsg fvts fnts (Apply f m) = showFn fnts (showMsg fvts fnts) (Apply f m)

message as seen by sender
showSenderMsg = showMsg . senderMsg

>showSenderMsg :: VarTypeList -> FnTypeList -> Msg -> String
>showSenderMsg fvts fnts (Atom v) = 
>  if findtype1 fvts v == "TimeStamp" then "Timestamp."++v else v
>showSenderMsg fvts fnts (Encrypt k ms) = 
>  "Encrypt.(" ++ showSenderMsg fvts fnts k ++ ", <" ++ 
>  commaConcat (map (showSenderMsg fvts fnts) ms) ++ ">)"
>showSenderMsg fvts fnts (Sq ms) = 
>  "Sq.<" ++ commaConcat (map (showSenderMsg fvts fnts) ms) ++ ">"
>showSenderMsg fvts fnts (Xor m m') = 
>  "Xor.(" ++ showSenderMsg fvts fnts m ++ ", " ++ 
>  showSenderMsg fvts fnts m' ++ ")"
>showSenderMsg fvts fnts (Undec m v) = showSenderMsg fvts fnts m
>showSenderMsg fvts fnts (Forwd v m) = v
>showSenderMsg fvts fnts (Apply f m) = 
>  showFn fnts (showSenderMsg fvts fnts) (Apply f m)

Message as seen by receiver; used to produce definition of agent process.

>showReceiverMsg :: 
>  VarTypeList -> FnTypeList -> InverseKeyList -> Msg -> String
>showReceiverMsg fvts _ _ (Atom v) = 
>  if findtype1 fvts v == "TimeStamp" then "Timestamp."++v else v
>showReceiverMsg fvts fnts fiks (Encrypt k ms) =
>  "Encrypt.("++showReceiverKey fvts fnts fiks k++", <" ++ 
>  commaConcat (map (showReceiverMsg fvts fnts fiks) ms) ++ ">)"
>showReceiverMsg fvts fnts fiks (Sq ms) = 
>  "Sq.<" ++ commaConcat (map (showReceiverMsg fvts fnts fiks) ms) ++ ">"
>showReceiverMsg fvts fnts fiks (Xor m m') = 
>  "Xor.(" ++ showReceiverMsg fvts fnts fiks m ++ ", " 
>          ++ showReceiverMsg fvts fnts fiks m' ++ ")"
>showReceiverMsg fvts fnts fiks (Undec m v) = v
>showReceiverMsg fvts fnts fiks (Forwd v m) = showReceiverMsg fvts fnts fiks m
>showReceiverMsg fvts fnts fiks (Apply f m) = 
>  showFn fnts (showReceiverMsg fvts fnts fiks) (Apply f m)

Message as seen by receiver; used to produce definition of INPUT_INT_MSG

>showReceiverChanMsg :: VarTypeList -> FnTypeList -> Msg -> String
>showReceiverChanMsg fvts _ (Atom v) = 
>  if findtype1 fvts v == "TimeStamp" then "Timestamp."++v else v
>showReceiverChanMsg fvts fnts (Encrypt k ms) =
>  "Encrypt.("++showReceiverChanMsg fvts fnts k++", <" ++ 
>  commaConcat (map (showReceiverChanMsg fvts fnts) ms) ++ ">)"
>showReceiverChanMsg fvts fnts (Sq ms) = 
>  "Sq.<" ++ commaConcat (map (showReceiverChanMsg fvts fnts) ms) ++ ">"
>showReceiverChanMsg fvts fnts (Xor m m') = 
>  "Xor.(" ++ showReceiverChanMsg fvts fnts m ++ ", " 
>          ++ showReceiverChanMsg fvts fnts m' ++ ")"
>showReceiverChanMsg fvts fnts (Undec m v) = v
>showReceiverChanMsg fvts fnts (Forwd v m) = showReceiverChanMsg fvts fnts m
>showReceiverChanMsg fvts fnts (Apply f m) = 
>  showFn fnts (showReceiverChanMsg fvts fnts) (Apply f m)

Show key, as seen by receiver

>showReceiverKey fvts fnts fiks (Atom k) = 
>  "inverse(" ++ inverse1 fiks k ++ ")"
>showReceiverKey fvts fnts fiks (Apply f as) =
>  showReceiverMsg fvts fnts fiks (Apply f as)
>showReceiverKey fvts fnts fiks (Undec m k) =
>  "inverse(" ++ inverse1 fiks k ++ ")"
>showReceiverKey fvts fnts fiks (Forwd k m) =
>  showReceiverKey fvts fnts fiks m


show message, using _ for all fields except first occurance of pat (agent
name) if flag==false, for all fields if flag==true

>showPatMsg :: VarName -> Bool -> Msg -> (Bool,String)
>showPatMsg pat b (Atom v) 
>  | v == pat && not b   = (True,v)
>  | otherwise           = (b,"_")
>showPatMsg pat b (Encrypt k ms) = 
>  (b', "Encrypt.(_, <" ++ commaConcat sts ++ ">)")
>  where -- (b',kst) = showPatMsg pat b k
>        (b',sts) = showPatSeq pat b ms
>showPatMsg pat b (Sq ms) = 
>  (b', "Sq.<" ++ commaConcat sts ++ ">")
>  where (b',sts) = showPatSeq pat b ms
>showPatMsg pat b (Xor m m') = (b'', "Xor.("++sts++", "++sts'++")")
>  where (b',sts) = showPatMsg pat b m
>        (b'',sts') = showPatMsg pat b' m'
>showPatMsg pat b (Undec m v) = (b,"_")
>showPatMsg pat b (Forwd v m) = (b,"_")
>showPatMsg pat b (Apply f as) = (b, "_") --  f++"("++commaConcat sts++")")
>  -- where (b',sts) = showPatSeq pat b (map Atom as)

>showPatSeq :: VarName -> Bool -> [Msg] -> (Bool,[String])
>showPatSeq pat b [] = (b,[])
>showPatSeq pat b (m:ms) = (b'', st:sts)
>  where (b',st) = showPatMsg pat b m
>        (b'',sts) = showPatSeq pat b' ms


============================================

components of message reachable without doing any decryption, as seen by
receiver.

>components1 (Atom a) = [Atom a]
>components1 (Sq ms) = ms
>components1 (Encrypt k ms) = [Encrypt k ms]
>components1 (Xor m m') = [Xor m m']
>components1 (Apply f ms) = [Apply f ms]
> -- components1 (Undecryptable k ms v) = [Atom v]
> -- components1 (Forward v k ms) = [Encrypt k (map receiverMsg ms)]

Print set giving type of message

>typeSet :: VarTypeList -> FnTypeList -> DataTypeDefs -> VarName -> Msg -> String
>typeSet fvts fnts dtdefs v m = 
>  case [t | (v',t) <- fvts, v==v'] of
>    [t] -> "addGarbage_("++t++")"
>    []  -> "addGarbage_({" ++ showMsg fvts fnts (varMsg m) ++ " | " ++ 
>           commaConcat (makeVarGens fvts fnts dtdefs (varFields m)) ++ "})"

>makegens fvts dtdefs = 
>  commaConcat . filter (/= "") . map (makegen fvts dtdefs)

>makegen fvts dtdefs v = 
>  let dtcons0 = [cons | (_,pats,_) <- dtdefs, (cons,[]) <- pats]
>  in
>  if member v dtcons0  -- if v is a datatype constructor...
>  then ""              -- then produce no generator
>  else if take 10 v == "Timestamp." then drop 10 v++" <- TS"
>  else v ++ " <- " ++ findtype1T fvts v

>makeVarGens fvts fnts dtdefs args = 
>  filter (/= "") (map (makeVarGen fvts fnts dtdefs) args)

>makeVarGen fvts fnts dtdefs (Simple v) = 
>  let dtcons0 = [cons | (_,pats,_) <- dtdefs, (cons,[]) <- pats]
>  in
>  if member v dtcons0  -- if v is a datatype constructor...
>  then ""              -- then produce no generator
>  else if take 10 v == "Timestamp." then drop 10 v++" <- TS"
>  else v ++ " <- " ++ findtype1T fvts v
>makeVarGen fvts fnts dtdefs (Subst v m) = 
>  v ++ " <- " ++ typeSet fvts fnts dtdefs v m

