Parser, to take string representing file, and return corresponding
representation of protocol, of type PreInput

>module Parse (parse) where

>import Useful
>import Maybe1
>import Pprint
>import Atoms
>import Messages
>import Types
>import GenParse
>import MiscParse
>import Msgparse
>import Parse1

>type FnType1 = ([TypeName], TypeName)
>type FnTypeList1 = [(VarName, FnType1)]

==========================================================================

The main parsing function

>parse :: String -> Maybe_ Input
>parse file = 
>  let sects = (map (\(s,ls) -> (getKeyword s, ls)) . split . 
>               filter (not . comment) . makeLogicalLines . lines)
>                  file
>      badKWerrs =
>        ["Unrecognized keyword: " ++ st ++ "\n" | (None st,_) <- sects]
>      duplicates [] = []
>      duplicates (kw : kws) 
>        | member kw kws      = kw : duplicates kws
>        | otherwise          = duplicates kws
>      duplicateKWerrs = 
>       ["Duplicated keyword: " ++ show kw ++ "\n" | 
>           kw <- duplicates (map fst sects)]
>      warnings = concat (badKWerrs ++ duplicateKWerrs)
>  in 
>  if warnings /= "" then Error warnings 
>  else
>  let mfvinfo = parseFVars (select FreeVars sects)
>      mprocs = parseProcsInfo (select Processes sects)
>      mprotdesc = parseProtDesc (select ProtDesc sects)
>      --mspecs = parseSpecs (select Specs sects)
>      --mvinfo = parseVars (select ActualVars sects)
>      minlines = parseFunctions (select Inlines sects)	
>      external = parseExternal (select External sects)
>      --mactAgents = parseActualAgents (select System sects)
>      --mintInfo = parseIntruderDef (select IntruderInf sects)
>      --mequivs = parseEquivs (select Equivs sects)
>      --mchannels = parseChannels (select Channels sects)
>      output1 = 
>         (mfvinfo &&& mprocs &&& mprotdesc &&& minlines)

--mspecs &&& mvinfo &&& 
--&&& mactAgents &&& mintInfo &&& mequivs &&& mchannels

>  in 
>  if isError output1 then (let Error e = output1 in Error e) 
>                          -- type coercion hack
>  else
>  let Yes (fvts,fnts,fiks,dtdefs) = mfvinfo
>      Yes procs = mprocs
>      Yes protdesc = mprotdesc
>      --Yes (secrets,auths) = mspecs
>      --Yes (vts,iks,tsinfo) = mvinfo
>      Yes inlines = minlines
>      --Yes (actAgents,withdraw) = mactAgents
>      fnts' = [(f, if member (Symbolic f) inlines then Symb dts rt 
>                   else Defed dts rt) | 
>                 (f,(dts,rt)) <- fnts] ++
>              [(f, HashFunction (take 3 t)) | (f,t) <- fvts, (drop 3 t)=="HashFunction"] ++
>              -- Turn datatype constructors into functions:
>              [(cons, Symb args tn) | 
>                 (tn,pats,_) <- dtdefs, (cons, args) <- pats, args/=[]]
>      fvts' = filter ((/= "HashFunction") . snd) fvts ++
>              -- turn datatype constructors with no arguments into variables:
>              [(cons, tn) | (tn,pats,_) <- dtdefs, (cons, []) <- pats]
>      --Yes ((((((intruderId,ik0),intruderProcs),crackables),deductions),
>      --       guessables),stale) = mintInfo
>      --Yes equivs = mequivs
>      --Yes channels = mchannels
>  in Yes (fvts', fnts', fiks, dtdefs, procs, protdesc, --secrets, auths, 
>          --vts, iks, tsinfo, 
>	   inlines,external) 
>	   --actAgents, withdraw, stale, 
>	   --intruderId, ik0, intruderProcs, crackables, deductions, 
>          --guessables, equivs, channels

========================================================================
Recognize section headers, and convert into keywords

>data Keyword = FreeVars | Processes | ProtDesc | Specs | ActualVars |
>               Inlines | System | IntruderInf | Equivs | Channels |
>               Preamble | None String | External
>               deriving (Eq, Show)

>getKeyword :: Line -> Keyword
>getKeyword "#Preamble" = Preamble
>getKeyword "#Freevariables" = FreeVars
>getKeyword "#Processes" = Processes
>getKeyword "#Protocoldescription" = ProtDesc
>getKeyword "#Specification" = Specs
>getKeyword "#Actualvariables" = ActualVars
>getKeyword "#Inlinefunctions" = Inlines
>getKeyword "#Functions" = Inlines
>getKeyword "#System" = System
>getKeyword "#IntruderInformation" = IntruderInf
>getKeyword "#Equivalences" = Equivs
>getKeyword "#Channels" = Channels
>getKeyword "#External" = External
>getKeyword other = None other

==========================================================================
Parse #ActualVariables section

>parseVars :: [Line] -> Maybe_ (ActVarTypeList, InverseKeyList, TimeStampInfo)
>parseVars ls = 
>  let	ls' = map dropWhite ls
>	ls'' = -- lines with variable declarations
>              [l | l <- ls', -- tok = hd(tokenize l),
>                  hd(tokenize l) /= "InverseKeys" 
>                  && hd(tokenize l) /= "TimeStamp" 
>                  && hd(tokenize l) /= "MaxRunTime"]
>	ikls = [l | l <- ls', take 11 l == "InverseKeys"]
>	tsls = [(l, tokenize l) | l <- ls', take 9 l == "TimeStamp"]
>	rtls = [(l, tokenize l) | l <- ls', take 10 l == "MaxRunTime"]
>	maybevars = concat @@ 
>                  combineMaybes 
>                    [(combineMaybes . map checkVar) ## getactvars l | l <- ls'']
>	checkVar (v, AtomType t, s) = 
>		case s of
>		 Error_Status -> Error "Unknown Type Status"
>		 otherwise -> Yes (v, t, s)
>	checkVar (v, FnType _ _, _) = Error "FnType"
>  in
>  (\ ((vl,il),tsi) -> (vl,il,tsi)) @@ 
>     ((maybevars &&& getinv ikls) &&& gettsinfo tsls rtls)
>    

Time stamp definition

>gettsinfo :: [(Line, [String])] -> [(Line, [String])] -> Maybe_ TimeStampInfo
>gettsinfo [] [] = Yes NotUsed
>gettsinfo [] rtls = (\rt -> MRTUsed rt) @@ getrt rtls
> -- Error "Unnecessary MaxRunTime definition.\n"
>gettsinfo [(l, ["TimeStamp", "=", t0, ".", ".", t1])] rtls 
>  | isNum t0 && isNum t1 
>     = (\rt -> TimeStampsUsed (parseInt t0) (parseInt t1) rt) @@ getrt rtls
>gettsinfo [(l,_)] _ = Error ("Bad TimeStamp definition: "++l++"\n")
>gettsinfo _ _ = Error ("Multiple TimeStamp definitions\n")

>getrt :: [(Line, [String])] -> Maybe_ Int
>getrt [] = Error "No MaxRunTime definition"
>getrt [(l, ["MaxRunTime", "=", rt])] | isNum rt = Yes (parseInt rt)
>getrt [(l,_)] = Error ("Bad MaxRunTime definition: "++l++"\n")
>getrt _ = Error ("Multiple MaxRunTime definitions\n")


>parseExternal :: [Line] -> ExternalList
>parseExternal []=[]
>parseExternal (t:q)=((parseExt t):(parseExternal q))

>parseExt :: Line -> External
>parseExt l = let l2=removeSpace l in
>     let (a,b)=parseNom l2 in
>        let (c,d)=parseNom b in
>            extern (a,c)

(t:q)=((removeSpace t):(parseExternal q))

==========================================================================
Parse #Functions section

>parseFunctions :: [Line] -> Maybe_ [Function]
>parseFunctions ls = concat @@ (combineMaybes (map parseFunction ls))

>parseFunction :: Line -> Maybe_ [Function]
>parseFunction [] = Error "Parse error in function definition: Empty line"
>parseFunction l = let l2=removeSpace l in
>     let (a,b)=parseNom l2 in
>        let (c,d)=parseNom b in
>           Yes [Defined (a,[],c)]

>parseNom :: String -> (String,String)
>parseNom [] = ([],[])
>parseNom (':':q)=([],q)
>parseNom ('=':q)=([],q)
>parseNom (t:q)  =let (a,b)=parseNom q in ((t:a),b)

>removeSpace :: String -> String
>removeSpace []=[]
>removeSpace (' ':q)=removeSpace q
>removeSpace (t:q)  =(t:(removeSpace q))

parseFunction = topLevel1 function "Parse error in function definition:"


inline ::= fnname "(" comma_list ")" "=" inlineRHS | symbolic comma_list
inlineRHS ::= varname -- only atoms at present

>function :: Parse Char [Function]
>function = 
>  word1ws "symbolic" ^^> comma_list >>> (\ fs -> map Symbolic fs)
>  |||
>  word1ws "inline" ^^> functionLHS ^^^ tokenws '=' ^^> many anyc
>    >>> (\ ((f, as), rhs) -> [Inline (f, as, rhs)])
>  |||
>  functionLHS ^^^ tokenws '=' ^^> fnRHS
>    >>> (\ ((f, as), rhs) -> [Defined (f, as, rhs)])

>anyc = spot (\ _ -> True)

>functionLHS = word ^^^ tokenws '(' ^^> fn_args

>fn_args :: Parse Char [String]
>fn_args = listOf fn_arg ',' <^^ tokenws ')'

>fn_arg = wordws ||| word1ws "_"

>fnRHS :: Parse Char FunctionRHS
>fnRHS = wordws

fn_arg ^^^ fn_args' >>> cons 
fn_args' =
  tokenws ',' ^^> fn_arg ^^^ fn_args' >>> cons
  |||
  tokenws ')' >>> (\ _ -> []) -}

==========================================================================

Parse #System section

*** Rewrite to parse sequential compositions

>parseActualAgents :: [Line] -> Maybe_ (ActualAgents, Bool)
>parseActualAgents ls =
>  let	ls'  = map dropWhite ls
>	wdls = [l | l<-ls', hd(tokenize l)=="WithdrawOption"]
>	wdres = if length(wdls)==1
>		then 
>			withdrawOption (hd(wdls))
>		else
>			if (wdls == [])
>			then	
>				Yes False
>			else
>				Error "Multiple Withdraw Options Defined."
>	ls'' = [l | l<-ls', hd(tokenize l)/="WithdrawOption"]
>  in
>	parseActualAgents1 ls'' &&& wdres 

>parseActualAgents1 :: [Line] -> Maybe_ ActualAgents
>parseActualAgents1 ls = 
>	combineMaybes (map parseAgent ls)

>parseAgent = topLevel1 parseAgent1 "Bad agent specification"

Parse possibly sequential agent, eg:

  INITIATOR(Alice, Na1) ; INITIATOR(Alice, Na2)

or starred form, eg:

  INITIATOR(Alice, Na) *

>parseAgent1 :: Parse Char SeqActualAgent
>parseAgent1 = 
>  parseOneAgent <^^ word1ws "*" >>> (\p -> Star p)
>  |||
>  listOf (ws parseOneAgent) ';' >>> (\ps -> SeqComp ps)

parse single instance of agent (eg "INITIATOR(Alice, Na)")

>parseOneAgent :: Parse Char ActualAgent
>parseOneAgent = word ^^^ token '(' ^^> termed_comma_list ')'

>withdrawOption :: Line -> Maybe_ Bool
>withdrawOption = 
>  topLevel1 withdrawOption1 "Error in Withdraw Option:"

>withdrawOption1 :: Parse Char Bool
>withdrawOption1 = 
>  word1ws "WithdrawOption" ^^> word1ws "=" ^^> 
>  ((word1ws "True" >>> \_ -> True)
>  |||
>  (word1ws "False" >>> \_ -> False))


==========================================================================
Parse definition of intruder

intruderDef ::= intruderId intruderKnowledge
                [processInfo] [crackableInfo] [deductions]

>parseIntruderDef :: [Line] -> 
>			Maybe_ ((((((IntruderId,IntruderInitKnowledge), 
>				[ProcessName]),[CrackInfo]), 
>				[Deduction]), [TypeName]), Bool)
>parseIntruderDef (l:l':ls) = 
>  let
>	ls'  = map dropWhite ls
>	stalels = [l | l<-ls', hd(tokenize l)=="StaleKnowledge"]
>	staleL = if length(stalels)==1
>		then 
>			staleKnowledge (hd(stalels))
>		else
>			if (stalels == [])
>			then	
>			    Yes True
>			else
>			    Error "Multiple StaleKnowledge Options Defined."
>	ls'' = [l | l<-ls', hd(tokenize l)/="StaleKnowledge"]
>  in
>	intruderId l &&& intruderKnowledge l' &&& processInfo ls''
>	        &&& crackables ls'' &&& deductions ls'' &&& guessables ls'' &&& staleL

------------------ Stale Knowledge

>staleKnowledge :: Line -> Maybe_ Bool
>staleKnowledge = 
>  topLevel1 staleKnowledge1 "Error in Stale Knowledge Flag:"

>staleKnowledge1 :: Parse Char Bool
>staleKnowledge1 = 
>  word1ws "StaleKnowledge" ^^> word1ws "=" ^^> 
>  ((word1 "True" >>> \_ -> True)
>  |||
>  (word1 "False" >>> \_ -> False))

------------------- Intruder's identity

IntruderId ::= "Intruder" "=" String

>intruderId :: Line -> Maybe_ IntruderId
>intruderId = 
>  topLevel1 intruderId1 "Error in intruder identity:"

>intruderId1 :: Parse Char IntruderId
>intruderId1 = 
>  word1 "Intruder" ^^> tokenws '=' ^^> wordws


----------------- Intruder's initial knowledge

intruderKnowledge  ::= "IntruderKnowledge" "=" "{" String intruderKnowledge'
intruderKnowledge' ::= [, String] "}"

>intruderKnowledge :: Line -> Maybe_ IntruderInitKnowledge
>intruderKnowledge = 
>  topLevel1 intruderKnowledge1 "Error in intruder knowledge:"

>intruderKnowledge1 :: Parse Char IntruderInitKnowledge
>intruderKnowledge1 =
>  word1 "IntruderKnowledge" ^^> tokenws '=' ^^> tokenws '{' 
>    ^^> intuderKnowledgeDatum ^^^ intruderKnowledge2
>  >>> cons
>  -- (\ (m, ms) -> m:ms)

>intruderKnowledge2 =
>  tokenws ',' ^^> intuderKnowledgeDatum ^^^ intruderKnowledge2 
>     >>> cons
>  |||
>  tokenws '}' >>> (\ _ -> [])

>intuderKnowledgeDatum :: Parse Char Msg
>intuderKnowledgeDatum = cpt -- fnappOrVar

--------------- Processes to be incorporated into the intruder

>processInfo :: [Line] -> Maybe_ [ProcessName]
>processInfo ls = 
>  let	ls'  = map dropWhite ls
>	ls'' = [l | l<-ls', hd(tokenize l)=="IntruderProcesses"]
>	ls'''= if (ls'' == []) then [] else hd(ls'')
>  in
>	intruderProcesses (ls''')

>intruderProcesses :: Line -> Maybe_ [ProcessName]
>intruderProcesses = 
>  topLevel1 intruderProcesses1 "Error in IntruderProcesses:"

>intruderProcesses1 :: Parse Char [ProcessName]
>intruderProcesses1 =
>  word1 "IntruderProcesses" ^^> tokenws '=' 
>    ^^> wordws ^^^ intruderProcesses2
>  >>> cons
>  |||
>  succeed []

>intruderProcesses2 =
>  tokenws ',' ^^> wordws ^^^ intruderProcesses2 
>     >>> cons
>  |||
>  succeed []

------------ Types crackable by intruder

>crackables :: [Line] -> Maybe_ [CrackInfo]
>crackables ls = 
>  let	ls' = [l | l<-ls, hd(tokenize l)=="Crackable"]
>  in if ls'==[] then Yes [] else crackables2 (hd ls')

>crackables2 = topLevel1 crackables3 "Error in crackables"

>crackables3 = word1 "Crackable" ^^> tokenws '=' ^^> listOf crackItem ','

>crackItem :: Parse Char CrackInfo
>crackItem = wordws ^^^ crackTime

>crackTime :: Parse Char (Maybe Int)
>crackTime = 
>  word1ws "(" ^^> number <^^ word1ws ")" >>> Just
>  ||| succeed Nothing

--------------- Guessable values

>guessables :: [Line] -> Maybe_ [TypeName]
>guessables ls = 
>  let	ls' = [l | l<-ls, hd(tokenize l)=="Guessable"]
>  in if ls'==[] then Yes [] else guessable2 (hd ls')

>guessable2 = topLevel1 guessable3 "Error in guessables"

>guessable3 = word1 "Guessable" ^^> tokenws '=' ^^> listOf wordws ','

--------------- Extra deductions of form

>testded = "forall ka, kb : Key . ka, {kb}{Ks} |- {kb}{ka}"

deduction ::= "forall" quants Msg ["," Msg] "|-" Msg

>deductions :: [Line] -> Maybe_ [Deduction]
>deductions ls = 
>  let	ls' = map dropWhite ls
>	ls''= [l | l<-ls', 
>                  let w1 = hd(tokenize l) 
>                  in w1 `notElem` ["IntruderProcesses", "IntruderRoles", 
>                                   "Crackable", "Guessable"]]
>  in
>	combineMaybes (map deduction ls'')

>deduction = topLevel1 deduction1 "Error in deduction"

>deduction1 :: Parse Char Deduction
>deduction1 =
>  word1ws "forall" ^^> conds ^^^ parsemsg ^^^ deduction2
>    >>> (\ (qs, (m, (ms, m'))) -> (qs, flatten m++ms, m'))

>deduction2 :: Parse Char ([Msg], Msg)
>deduction2 = 
>  tokenws ',' ^^> parsemsg ^^^ deduction2 >>> 
>      (\ (m, (ms, m')) -> (flatten m++ms, m'))
>  |||
>  word1ws "|-" ^^> parsemsg >>> (\ m -> ([], m))

Flatten sequences, to recognize comma-separated messages in deductions as
individual messages, not sequences.

>flatten (Sq ms) = ms
>flatten m = [m]

Parse conditions

conds ::= cond [";" cond] "." 

>conds :: Parse Char [Cond]
>conds = cond ^^^ conds' >>> (\(c,cs) -> c++cs)

>conds' :: Parse Char [Cond]
>conds' =
>  word1ws ";" ^^> cond ^^^ conds' >>> (\(c,cs) -> c++cs)
>  |||
>  word1ws "." >>> (\_ -> [])

Parse condition, firstly storing result as Quant, then translating into
Cond:

cond = varName {, varName} [: typename1]

>type Quant = ([VarName], TypeName)
>quant2conds (vs,t) = [(v,t) | v <- vs]

>cond :: Parse Char [Cond]
>cond = 
>  word ^^^ cond' >>> (\ (v, (vs,typename)) -> quant2conds(v:vs, typename))

>cond' :: Parse Char Quant
>cond' = 
>  tokenws ',' ^^> word ^^^ cond' 
>    >>> (\ (v, (vs, typename)) -> (v:vs, typename))
>  |||
>  tokenws ':' ^^> typename1 >>> (\ typename -> ([], typename))
>  |||
>  succeed ([], "")

typename1 ::= "Message" | typename
The following parser could be simplified, but this makes it explicit that 
Message is distinguished.

>typename1 =
>  word1ws "Message" ||| word 

Previously Message* was also distinguished

>  -- word1ws "Message" ^^^ word1ws "*" >>> (\_ -> "Message*")

=============================================================
Parse equivalences

>parseEquivs :: [Line] -> Maybe_ [Equivalence]
>parseEquivs ls = combineMaybes (map equiv ls)

>equiv = topLevel1 equiv1 "Error in equivalence"

equiv1 ::= forall cond {; cond} "." Msg "=" Msg
 
>equiv1 :: Parse Char Equivalence
>equiv1 = 
>  word1ws "forall" ^^> conds ^^^ parsemsg ^^^ word1ws "=" ^^> parsemsg 
>    >>> (\ (cs, (m1, m2)) -> (cs, m1, m2))

==================================================================
Parse channel information

>parseChannels :: [Line] -> Maybe_ ChannelInfo
>parseChannels [] = Yes (False, False, False)
>parseChannels ls = combineCIs @@ combineMaybes (map chan ls)

>chan = topLevel1 chan1 "Error in channel information"

>chan1 :: Parse Char ChannelInfo
>chan1 = 
>  word1ws "secret" ^^> succeed (True,False,False)
>  ||| 
>  word1ws "authenticated" ^^> succeed (False,True,False)
>  ||| 
>  word1ws "direct" ^^> succeed (False,False,True)

Combine list of ChannelInfo tuples into single tuple

>combineCIs = addDirect . foldr f (False,False,False)
>  where f (b1,b2,b3) (b1',b2',b3') = (b1||b1', b2||b2', b3||b3')
>        addDirect (b1,b2,b3) = (b1,b2,b1||b2||b3)
