Checking, particularly type checking

>module TypeCheckDI (typecheckErrors_DI) where

>import Maybe1
>import Useful
>import Pprint
>import Atoms
>import Messages
>import Types
>import Accumulation

>type Warning = String
>type Error = String

Perform various checks on the input

>typecheckErrors_DI fvts procs protdesc secrets auths actvts 
>	timestampInfo actualAgents withdraw stale intruderId 
>	intruderInitKnows intruderProcs equivs = 
>  let	vts = [(v,t) | (v,t,_) <- actvts]
>	generations = concat[gens | (_,_,_,_,gens) <- procs]
>	existGenerations = generations/=[]
>	dit = [t | (_,t,s) <- actvts, s /= Normal]
>       isDit v = elem (findtype1 fvts v) dit
>	generates = concat[gens | (_,_,_,_,gens) <- procs]
>       internalAgents = [id | (id,name,_,_,_) <- procs, 
>                               member name intruderProcs]
>       accumList = accumulation fvts protdesc dit
>       diAccumList0 = [(nr, (a1, filter isDit vs1), (a2, filter isDit vs2))  |
>                        (nr, (a1,vs1), (a2,vs2)) <- accumList]
>       diAccumList = [(nr, (a1, vs1), (a2, vs2)) | 
>                        (nr, (a1, vs1), (a2, vs2)) <- diAccumList0, 
>                        vs1/=[] || vs2/=[]]
>
>	-- used in 4. below
>	nonAtomicGens = [v | v <- generates, not(atomMember v fvts)]
>
>	-- used in 5. below
>	multiGens = listMultiples generates
>
>	-- used in 6. below
>	argsGenerates = [(id,intersection arg gen) | 
>				(_,id,arg,_,gen) <- procs]
>	-- used in 8a) below
>	reqForeground = remdups(insufficientForegrounds fvts 
>					generates actvts dit)
>	-- used in 8b)  below
>	diTypesInGens = [findtype1 fvts v | v <- generates]
>	reqKnownBack = [t | t <- diTypesInGens, 
>			 not(existKnownBack t actvts) &&
>				t/="_Internal"]
>
>	-- used in 8c) below
>	multiBack = [t | (_,t,s) <- actvts, checkMultiBack t actvts]
>
>	-- used in 9. below
>	generatesTypes = remdups(concat[map (findtype1 fvts) gens |
>				(_,_,_,_,gens) <- procs])
>	checkDiActVars = dit\\generatesTypes
>
>	-- used in 10. below
>	diNotInIntruder = [v | (v,t,s) <- actvts, s==KnownBackground,
>				not(member (Atom v) intruderInitKnows)]
>	-- used in 11. below
>	extIntProcs = intersection intruderProcs (gatherAgents actualAgents)
>
>	-- used in 12. below
>	checkIntruderProcs = intruderProcs\\([name | (_,name,_,_,_)<-procs])
>
>	-- used in 13. below
>	noInstanceProcs = [id | id <- intruderProcs, 
>					noInstance id actvts] 
>	-- used in 14. below
>	multiSeqInstances = [ls | SeqComp ls <- actualAgents, 
>					(length ls) > 1]
>	-- used in 15. below
>	multiParInstances = listMultiples(gatherAgents actualAgents)
>
>	-- used in 16. below
>	errors = checkGenExists generates protdesc
>
>	-- used in 17. below
>	agentsWithIntrType = [v | (v,t) <- vts,
>				  v /= intruderId,
>				  t == findtype1 vts intruderId]
>
>	-- used in 18(a). below
>	simpleCase1_violation = 
>		[nr | (nr,(s,sl),_) <- diAccumList, 
>				member s internalAgents,
>					length(sl) > 1]
>
>	-- used in 18(b). below
>	simpleCase2_violation = simpleCase2_check protdesc diAccumList
>					internalAgents
>	-- used in 19. below
>	internalProcType = checkInternalProcType fvts vts intruderId 
>						   procs intruderProcs
>	-- used in 20. below
>	internalAuthAgents = [r | (r,_,_) <- auths, 
>					member r internalAgents] ++
>			     [c | (_,c,_) <- auths, 
>					member c internalAgents]
>	secrecyAgents = concat([(a:ls) | Sec a _ ls <- secrets] ++
>			       [(a:ls) | StrongSec a _ ls <- secrets])
>	internalSecAgents = [id | id <- secrecyAgents, 
>					member id internalAgents]
>	internalSpecs = remdups(internalSecAgents ++ internalAuthAgents)
>	-- used in 22. below
>	actvtsTypes = remdups [t | (_,t,s) <- actvts, s/=Normal]
>	internalAgentsBanned = False
>  in
>	if existGenerations	-- then treat as di script.
>	then
>	  -- 1. No timestamps allowed in system.
>	  (if (timestampInfo /= NotUsed)
>	   then 
>		"No timestamps are not permitted within a script" ++ 
>		"using data independence.\n\n"
>	   else "") ++
>
>	  -- 2. No authentication specs involving timestamps.
>	  concat
>	   ["The timed authentication specification:" ++ show au 
>	    ++ "\n" ++ "is not permitted within a script using data "
>	    ++ "independence.\n\n" | (_,_,au) <- auths, isTimedAuth au]++
>
>	  -- 3. No algebraic equivalences.
>	  (if (equivs /= [])
>	   then 
>		"Algebraic equivalences are not permitted within "++
>		"a script using data independence.\n\n"
>	   else "") ++
>
>	  -- 4. Every variable in a generation list in procs 
>	  --	(# Processes) must be atomic (i.e. of the form Atom X
>	  --	in fvts).
>	  concat
>	   ["Undeclared atomic variable(s) in 'generates' list(s): " ++ 
>		commaConcat nonAtomicGens ++ ".\n\n" |
>			nonAtomicGens/=[]] ++ 
>
>	  -- 5. Each variable in generations is unique.
>	  concat
>	   ["Error: The following data independent variables " ++ 
>	    "(in generates): " ++ commaConcat multiGens ++ 
>	    " are declared more than once in #Processes.\n" | 
>		multiGens/=[]] ++
>
>	  -- 6. Variables in generations are not declared elsewhere in
>	  --	procs (# Processes).
>	  concat
>	   ["Error: In #Processes, process " ++ id ++ " has variables " 
>	    ++ commaConcat ls ++ "\n declared as both arguments" ++
>	    " and data independent variables (generates).\n" | 
>			(id,ls) <- argsGenerates, ls/=[]] ++
>
>	  -- 8. Each DI type T declared in actvts must conform to
>	  --	rules: 
>	  -- a) must have variables declared as type T(Foreground).
>	  -- b) must have exactly one variable declared as type
>	  --    T(KnownBackground).
>	  -- c) (optional) exactly one variable declared as type
>	  --	T(UnknownBackground).
>	  concat
>	   ["Error: The following data independent types require " ++
>	    "more Foreground values to be declared in # Actual " ++
>	    "Variables:\n" ++ spaces 5 ++ 
>	    commaConcat reqForeground ++ ".\n" | reqForeground/=[]] ++
>	  concat
>	   ["Error: The following data independent types require " ++
>	    "KnownBackground values to be declared in #Actual " ++
>	    "Variables:\n" ++
>	    spaces 5 ++ commaConcat reqKnownBack ++ ".\n\n" |
>			False && reqKnownBack/=[]] ++
>	  concat
>	   ["Error: The following data independent types have too " ++
>	    "many Background values declared each in # Actual " ++
>	    "Variables\n: " ++ spaces 5 ++ commaConcat multiBack ++
>	    ".  For each, there should be at most 1.\n\n" |
>				multiBack/=[]] ++
>
>	  -- 9. For every DI type T used in the declarations in
>	  --	actvts (# Actual Variables), there must exist at least
>	  --	1 variable v of type T in a generation list (in procs).
>	  concat
>	   ["Error: The following types have been declared as data "++
>	    "independent in #Actual Variables:\n" ++ 
>	    commaConcat checkDiActVars ++ ",\n" ++ 
>	    "whereas these types do not have corresponding " ++
>	    "'generated' variables in #Processes.\n\n"
>		 | checkDiActVars/=[]] ++
>
>	  -- 10. Every actual variable v declared as some DI type
>	  --	 T(KnownBackground) in actvts must also be a member of
>	  --	 intruderInitKnows (i.e. intruder's initial knowledge).
>	  concat
>	   ["Error: The following actual variables declared as " ++
>	    "KnownBackground values: " ++ commaConcat diNotInIntruder
>	    ++ ",\n" ++ 
>	    "should be declared within the intruder's initial " ++
>	    "knowledge.\n\n" | diNotInIntruder/=[]] ++
>
>	  -- 11. The intersection of IntruderProcesses and 
>	  --	 (map firstEl actualAgents) must be the empty list.
>	  --	 i.e. A process is either internal (within the
>	  --	 intruder) or external, but cannot be both. 
>	  concat
>	   ["Error: Processes " ++ commaConcat extIntProcs ++ 
>	    " are declared as\n" ++
>	    "both external (in #Actual System) and internal (in " ++
>	    "IntruderProcesses).\n\n" | extIntProcs /= []] ++
>
>	  -- 12. Each process in intruderProcess (i.e. internal
>	  --	 processes) must be declared in procs.
>	  concat
>	   ["Error: The following internal processes " ++
>	    "(in IntruderProcesses)\n are not declared in " ++
>	    "#Processes:\n" ++ commaConcat checkIntruderProcs ++ 
>	    ".\n\n" | checkIntruderProcs/=[]] ++
>
>	  -- 13. Each process P in intruderProcess must have a
>	  --	 corresponding identity of type P declared in actvts 
>	  --	(i.e.  #Actual Variables).
>	  concat
>	   ["Error: The following internally declared processes: " ++
>	    commaConcat noInstanceProcs ++ ", \n" ++
>	    "require an actual variable declaration each in " ++
>	    "#Actual Variables.\n\n" | noInstanceProcs/=[]] ++
>
>	  -- 14. Each instantiation in actualAgents (#System) must be
>	  --	 of the form:  SeqComp LS, where length(LS) = 1.
>	  concat
>	   ["Error: No sequential instantiation of processes in #System\n" ++
>	    "should be declared in a script using data independence,\n" ++
>	    "since each process will be allowed to perform unbounded\n"++
>	    "number of runs with a fresh supply of values.\n\n" |
>			multiSeqInstances/=[]] ++
>
>	  -- 15. For each process P declared in procs (#Processes),
>	  --	 there can exist at most 1 instantiation in
>	  --	 actualAgents (#System).  i.e. It is not possible to
>	  --	 declare multiple instances of 1 process type running
>	  --	 in parallel. 
>	  concat
>	   ["Error: No parallel instances of the same process in #System\n" ++
>	    "should be declared in a script using data independence.\n\n"
>			| multiParInstances/=[]] ++
>
>	  -- 16. All variables declared in generates lists should be
>	  --	 used in protocol description.
>	  concat
>	   ["Error:  The following list of data independent variables\n"++
>	    "(in generates list) do not exist in the protocol" ++
>	    " description:\n"++ spaces 5 ++ commaConcat errors ++ 
>	    ".\n\n" | errors/=[] && False] ++
>	  -- 17.
>	  concat
>	   ["Error:  There should be at least one actual variable\n"++
>	    "declared with the same type as the intruder's identity.\n"
>		      | False && agentsWithIntrType==[]] 
>	   ++
>	  -- 18. The internal processes must satisfy the SIMPLE CASE
>	  --	 conditions. 
>	  -- 18 a) Every message in "protdesc" introduces at most 1
>	  --	   fresh value by an internal agent.
>	  concat
>	   ["Error:  Each message can introduce at most 1 fresh " ++
>	    "variable\n" ++
>	    "by an internal agent.  The following messages violates " ++
>	    "this condition:\n" ++
>	    spaces 10 ++ "Message(s) " ++
>	    commaConcat simpleCase1_violation ++ ".\n\n" | 
>		simpleCase1_violation /= []] ++
>
>	  -- 18 b) Every internal agent A must satisfy precisely one
>	  --	   of the following conditions:  Either,
>	  --	   i) A introduces at most 1 fresh value per run, or
>	  --	  ii) all messages created by A never depend on A
>	  --	      receiving a msg M, where M contains a fresh
>	  --	      variable (of any di type) previously introduced
>	  --	      by A. 
>	  concat
>	   ["Error:  For each internal agent X, the following " ++
>	    "condition\n" ++
>	    "must be satisfied (Simple Case):  Either, i) X introduces" ++
>	    " at most\n" ++
>	    "1 fresh value per run, or ii) all messages created by X " ++
>	    "never\n" ++
>	    "depend on X receiving a msg M, where M contains a fresh " ++
>	    "variable\n" ++
>	    "(of any di type) previously introduced by X.  This does "++ 
>	    "not hold\n" ++
>	    "for the following internal process(es): " ++
>	    commaConcat simpleCase2_violation ++ ".\n\n" | 
>		simpleCase2_violation /= []] ++
>
>	  (if internalAgentsBanned then 
>	    -- 19. Banned internal agent types.
>	    concat
>	     ["Error: The processes: " ++ commaConcat internalProcType ++
>	      "\n" ++ "should not be declared as Internal Processes " ++
>	      "(within the intruder)\n" ++ "since the intruder can take"++
>	      "on this (these) role(s)." | internalProcType /= []] 
>	    ++
>	    -- 20. Banned internal agents (used in specs.)
>	    concat
>	    ["Error: Process identities bound to internally declared\n"
>	     ++ "processes must not be used for specification purposes.\n"
>	     ++ "In your script, the variables: " ++
>	     commaConcat internalSpecs ++ " in the specifications\n" ++
>	     "are bound to internally declared process(es).\n\n" |
>		  	internalSpecs /= []]
>	   else [])
>	else 
>	  -- The script is clearly not meant to use DI and therefore
>	  -- the following checks are used to ensure this.
>
>	  -- 21. No declared internal processes.
>	  concat
>	   ["Error: Processes can only be declared as internal\n" ++
>	    "(using IntruderProcesses = ...) when the script\n"++
>	    "is making use of the data independence techniques.\n\n" |
>			intruderProcs/=[]]++
>
>	  -- 22. No di types declared in actvts (Actual Variables).
>	  concat
>	   ["Error:  There are data independent types declared in\n"++
>	    "#Actual Variables, even though there is no request\n"++
>	    "for freshly generated variables in #Processes.\n\n" |
>			actvtsTypes /= []]


==========================================================================

>insufficientForegrounds _ _ _ [] = []
>insufficientForegrounds fvts generates actvts (d:dit) =
>  let	foreList = [v | (v,t,s) <- actvts, s==Foreground,
>		         t==d]
>	gens = [v | v <- generates, findtype1 fvts v == d]
>  in
>	if length(foreList)<length(gens)
>	then
>		d:insufficientForegrounds fvts generates actvts dit
>	else
>		insufficientForegrounds fvts generates actvts dit


------------------------------------------------------------------------
checkInternalProcType
------------------------------------------------------------------------

>checkInternalProcType _ _ _ _ [] = []
>checkInternalProcType fvts vts intruderId procs (name:intrProcs) =
>  let	intrType = findtype1 vts intruderId
>	procType = head([findtype1 fvts id | (id,n,_,_,_) <- procs,
>					       n==name])
>  in
>	if procType==intrType then
>	   name:checkInternalProcType fvts vts intruderId procs intrProcs
>	else
>	   checkInternalProcType fvts vts intruderId procs intrProcs


-------------------------------------------------------------------------

>gatherAgents :: [SeqActualAgent] -> [ProcessName]

>gatherAgents [] = []
>gatherAgents (a:as) = 
>	case a of
>	  SeqComp ((id,_):ls) -> id:gatherAgents as
>	  Star (id,_)  -> id:gatherAgents as


>checkGenExists :: [VarName] -> ProtDesc -> [VarName]

>checkGenExists gens [] = gens
>checkGenExists gens ((_,_,_,_,m,_):ms) =
>  let gens' = gens\\(atoms_all m)
>  in checkGenExists gens' ms


This function returns a list of pairs (id,args) where all the elements
in args have a type which is in the data independent list dit.

>checkArgsDI ::
>	[(VarName,TypeName)] ->
>	[TypeName] ->
>	[(ProcessName,[Argument])] ->
>	[(ProcessName,[Argument])]

>checkArgsDI _ _ [] = []
>checkArgsDI fvts dit ((id,args):xs) =
>  let	args' = [v | v <- args, member (findtype1 fvts v) dit]
>  in
>	if (args'==[]) then checkArgsDI fvts dit xs
>	else (id,args'):checkArgsDI fvts dit xs


>existForeground :: TypeName -> ActVarTypeList -> Bool

>existForeground _ [] = False
>existForeground d ((v,t,s):vts) =
>	(d==t && s==Foreground) || existForeground d vts


>existKnownBack :: TypeName -> ActVarTypeList -> Bool

>existKnownBack _ [] = False
>existKnownBack d ((v,t,s):vts) =
>	(d==t && s==KnownBackground) || existKnownBack d vts


>checkMultiBack ::  TypeName -> ActVarTypeList -> Bool

>checkMultiBack d actvts =
>  let	known = [v | (v,t,s) <- actvts, d==t, s==KnownBackground]
>	unknown = [v | (v,t,s) <- actvts, d==t, s==UnknownBackground]
>  in
>	if (length known > 1 || length unknown > 1) 
>	then True
>	else False


>noInstance :: ProcessName -> ActVarTypeList -> Bool

>noInstance _ [] = False
>noInstance id ((v,t,s):vts) =
>	(id==t) || noInstance id vts


>atomMember :: VarName -> VarTypeList -> Bool

>atomMember _ [] = False
>atomMember a ((v,t):fvts) =
>	(a==v) || atomMember a fvts


&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

Verifying that the protocol description satisfies the conditions
required for the SIMPLE CASE (concerning the internal agents).  The
following functions are used to calculate "simpleCase2_violation" -
namely to verify the 2nd condition required for the simple case.

2. For each internal agent A, the following condition must be
satisfied:

Either: 
a) A introduces at most 1 fresh variable in the whole protocol
run, or

b) msg sent by A never relies on A receiving a msg containing a fresh
value previously introduced by A.


>data Role = Sender [VarName] | Rec [VarName]
>              deriving (Eq, Show)


*********************************************************************
simpleCase2_check

returns a list of all internal agents that violate the 2nd condition
of the simple case requirements.
---------------------------------------------------------------------

>simpleCase2_check _ _ [] = []
>simpleCase2_check protdesc diAccumList (a:intAgents) =
>  let	picksList = [(nr,sl) | (nr,(s,sl),_) <- diAccumList, s==a]
>	completeList = gatherAllMsgs a picksList protdesc
>	isCorrect_a = (totFreshIntro picksList) <= 1
>	isCorrect_b = verify_cond_b a [] True completeList
>  in
>	if (isCorrect_a || isCorrect_b) then
>	-- internal agent a satisfies cond. 2 of the simple case.
>		simpleCase2_check protdesc diAccumList intAgents
>	else
>	-- a does not satisfy cond. 2 
>		a:simpleCase2_check protdesc diAccumList intAgents
>  where
>	totFreshIntro ls = sum [length sl | (n,sl) <- ls]



*********************************************************************
gatherAllMsgs

(called by simpleCase2_check)

For a given internal agent X,
	returns a list of tuples (nr, role) of all the messages
involving X in protdesc.  Furthermore, the parameter "role" provides
information about whether X is the sender or receiver of the message
nr.  If X is the Sender, then it stores the list of fresh variables
introduced in message nr (if none then this list is empty); if X is
the receiver, it stores the list of all variables it receives in
message nr.
---------------------------------------------------------------------

>gatherAllMsgs _ _ [] = []
>gatherAllMsgs a intPicks ((_,nr,s,r,msg,_):ms) =
>  if (Agent a == s) then
>	(let ls = [(n, Sender fs) | (n,fs) <- intPicks, n==nr] 
>	 in
>	   if ls /= [] then
>		  head(ls):gatherAllMsgs a intPicks ms
>	   else (nr,Sender []):gatherAllMsgs a intPicks ms)
>  else if (Agent a == r) then
>	(nr,Rec (atomsRec msg)):gatherAllMsgs a intPicks ms
>  else
>	gatherAllMsgs a intPicks ms


*********************************************************************
verify_cond_b

(called by simpleCase2_check)

returns True iff condition 2(ii) of the simple case is satisfied;
returns False otherwise (i.e. if counter-example is found.
---------------------------------------------------------------------

>verify_cond_b _ _ _ [] = True
>verify_cond_b agent freshSoFar noError ((nr, role):xs) =
>    case role of
>	Sender ls  ->	if not(noError) then False
>			else verify_cond_b agent (freshSoFar++ls) noError xs
>	Rec ls	   ->	if (returned freshSoFar ls) then 
>			  verify_cond_b agent freshSoFar False xs
>			else verify_cond_b agent freshSoFar noError xs
>  where
>	returned fresh ls = (intersection fresh ls) /= []


[("2",Rec ["B","na"]), ("3",Rec ["nb"]), 
 ("4",Sender ["kab"]), ("5",Sender ["kab2"])]


