package core;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.security.Key;

/*
 * Created on May 30, 2003
 */

/**
 * This class represents a SecretKeyStore. 
 * It is equivalent to a KeyStore except that instead of storing Public and Secret Keys, it stores Secret Keys. 
 * @author Xavier Didelot
 */
public class SecretKeyStore {

	/**
	 * File in which the SecretKeyStore is located
	 */
	private String file;

	/**
	 * Creates a new SecretKeyStore
	 * @param file file in which the SecretKeyStore is located
	 */
	public SecretKeyStore(String file) {
		super();
		this.file = file;
	}

	/**
	 * This method returns the secret key of a given Agent
	 * @param id id of the Agent whose Secrey Key we are looking for
	 * @return the secret key of a given Agent
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public Key getKey(String id) throws IOException, ClassNotFoundException {
		FileInputStream fis = new FileInputStream(file);
		ObjectInputStream s = new ObjectInputStream(fis);
		String str = "";
		Key key = null;
		while (id.compareTo(str) != 0) {
			str = (String) s.readObject();
			key = (Key) s.readObject();
		}
		s.close();
		fis.close();
		return key;
	}

	/**
	 * This method adds a given key to this SecretKeyStore
	 * @param id id of the Agent whose key belongs to
	 * @param key key to store
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public void addKey(String id, Key key)
		throws FileNotFoundException, IOException {
		FileOutputStream fos = new FileOutputStream(file, true);
		ObjectOutputStream s = new ObjectOutputStream(fos);
		s.writeObject(id);
		s.writeObject(key);
		s.close();
		fos.close();
	}

	/**
	 * This sample method shows how to create a new SecretKeyStore.
	 * Here we create a new SecretKeyStore containing a new DES key for agent a and a new DES key for Agent b.
	 * @param args
	 */
	public static void main(String[] args) {
		String a = "a";
		String b = "b";
		try {
			Key keya = MessageKey.createKey("DES");
			Key keyb = MessageKey.createKey("DES");
			FileOutputStream fos = new FileOutputStream("secretkeyfile");
			ObjectOutputStream s = new ObjectOutputStream(fos);
			s.writeObject(a);
			s.writeObject(keya);
			s.writeObject(b);
			s.writeObject(keyb);		
			s.close();
			fos.close();
		} catch (Exception e) {
			e.printStackTrace();
		}

	}
}
